﻿namespace InputValidationTesting.Helper {
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.ComponentModel;
    using System.Linq;
    using System.Linq.Expressions;
    using System.Reflection;

    using InputValidationTesting.ViewModel;

    public class ErrorValidationHelper<T>
            where T : CustomerEditViewModel {

        private readonly IDictionary<String, IList<String>> _errors = new Dictionary<string, IList<string>>();

        private readonly IDictionary<string, KeyValuePair<Func<T, object>, ValidationBaseAttribute[]>> _validators;

        private readonly T _viewModel;

        public ErrorValidationHelper(T viewModel) {
            if (viewModel == null) {
                throw new ArgumentNullException(nameof(viewModel));
            }

            _viewModel = viewModel;
            _validators = new Dictionary<string, KeyValuePair<Func<T, object>, ValidationBaseAttribute[]>>();
            foreach (PropertyInfo property in typeof(T).GetProperties()) {
                ValidationBaseAttribute[] validations = GetValidations(property);
                if (validations.Length > 0) {
                    _validators.Add(property.Name,
                                    new KeyValuePair<Func<T, object>, ValidationBaseAttribute[]>(CreateValueGetter(property), validations));
                    Validate(_viewModel, property.Name);
                }
            }

            viewModel.PropertyChanged += delegate(object sender, PropertyChangedEventArgs args) {
                                             Validate(viewModel, args.PropertyName);
                                         };
        }

        public bool HasErrors {
            get { return _errors.Count > 0; }
        }

        public void AddError(string propertyName, string error, bool isWarning) {
            if (!_errors.ContainsKey(propertyName))
                _errors[propertyName] = new List<string>();

            if (!_errors[propertyName].Contains(error)) {
                if (isWarning)
                    _errors[propertyName].Add(error);
                else
                    _errors[propertyName].Insert(0, error);
                RaiseErrorsChanged(propertyName);
            }
        }

        public IEnumerable GetErrors(string propertyName) {
            if (string.IsNullOrEmpty(propertyName) || !_errors.ContainsKey(propertyName))
                return null;
            return _errors[propertyName];
        }

        public void RaiseErrorsChanged(string propertyName) {
            _viewModel.InvokeErrorsChanged(new DataErrorsChangedEventArgs(propertyName));
        }

        public void RemoveError(string propertyName, string error) {
            if (_errors.ContainsKey(propertyName) && _errors[propertyName].Contains(error)) {
                _errors[propertyName].Remove(error);
                if (_errors[propertyName].Count == 0)
                    _errors.Remove(propertyName);
                RaiseErrorsChanged(propertyName);
            }
        }

        public void Validate(T source, string columnName) {
            KeyValuePair<Func<T, object>, ValidationBaseAttribute[]> validators;
            bool isWatched = _validators.TryGetValue(columnName, out validators);
            if (isWatched) {
                object value = validators.Key(source);
                ValidationBaseAttribute[] validationAttributes = validators.Value;

                IEnumerable<ValidationBaseAttribute> invalidAttributes = validationAttributes.Where(v => !v.IsValid(value));
                IEnumerable<ValidationBaseAttribute> validAttributes = validationAttributes.Where(v => v.IsValid(value));

                foreach (ValidationBaseAttribute validationAttribute in invalidAttributes) {
                    AddError(columnName, validationAttribute.GetErrorMessageString(), false);
                }

                foreach (ValidationBaseAttribute validationAttribute in validAttributes) {
                    RemoveError(columnName, validationAttribute.GetErrorMessageString());
                }
            }
        }

        private static Func<T, object> CreateValueGetter(PropertyInfo property) {
            ParameterExpression instance = Expression.Parameter(typeof(T), "i");
            UnaryExpression cast = Expression.TypeAs(Expression.Property(instance, property), typeof(object));
            return (Func<T, object>)Expression.Lambda(cast, instance).Compile();
        }

        private static ValidationBaseAttribute[] GetValidations(PropertyInfo property) {
            return (ValidationBaseAttribute[])property.GetCustomAttributes(typeof(ValidationBaseAttribute), true);
        }

    }
}